use anyhow::{bail, Context, Result};
use base64::Engine;
use clap::{Parser, Subcommand};
use serde::{Deserialize, Serialize};
use std::collections::BTreeSet;
use std::fmt::Display;
use std::str::FromStr;
use std::sync::Arc;
use std::{collections::BTreeMap, path::PathBuf};
use tokio::{
    io::{AsyncBufReadExt, AsyncWriteExt, BufReader},
    net::{UnixListener, UnixStream},
    sync::{self, broadcast, mpsc, RwLock},
};
#[allow(unused_imports)]
use tracing::{error, info, instrument, warn};

static LICENSE: &str = include_str!("../LICENSE");

#[derive(Debug, Serialize, Deserialize, PartialEq, Eq, PartialOrd, Ord, Clone)]
enum SockType {
    Udp,
    Tcp,
}

impl Display for SockType {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            SockType::Udp => f.write_str("UDP"),
            SockType::Tcp => f.write_str("TCP"),
        }
    }
}

#[derive(Debug, Serialize, Deserialize, PartialEq, Eq, PartialOrd, Ord, Clone)]
struct PortMapping {
    sock_type: SockType,
    from: u16,
    to: u16,
}

impl Display for PortMapping {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.write_fmt(format_args!(
            "{}->{}/{}",
            self.from, self.to, self.sock_type
        ))
    }
}

#[derive(Debug, Serialize, Deserialize, PartialEq, Eq, PartialOrd, Ord, Clone)]
struct Remote {
    host: String,
    port: u16,
    public_key: Option<[u8; 32]>,
    default_tunnel: Tunnel,
    tunnels: Vec<Tunnel>,
}

#[derive(Debug, Serialize, Deserialize, PartialEq, Eq, PartialOrd, Ord, Clone)]
struct Tunnel {
    mappings: BTreeMap<u32, PortMapping>,
}

#[derive(Debug, Serialize, Deserialize, Default, Clone)]
struct Config<'a> {
    remotes: BTreeMap<String, Remote>,
    #[serde(skip)] // (<remote_name> <tunnel idx> <mapping idx>)
    flat_mappings: BTreeSet<(String, usize, &'a PortMapping)>,
}

impl<'a> Config<'a> {
    fn update_mappings(&'a mut self) {
        self.flat_mappings.clear();
        for (name, rem) in self.remotes.iter() {
            for (tidx, tun) in rem.tunnels.iter().enumerate() {
                for mapping in tun.mappings.values() {
                    self.flat_mappings.insert((name.clone(), tidx, mapping));
                }
            }
        }
    }
}

#[derive(Debug)]
struct AppState {
    active_config: Arc<RwLock<Config<'static>>>,
    config: Arc<RwLock<Config<'static>>>,
    notify: sync::mpsc::Sender<()>, // Used to tell the ipc server to reload
}

#[derive(Debug, Serialize, Deserialize)]
enum ClientOp {
    Export,
    Import,
    Add(PortMapping),
    Del(Vec<usize>),
    Reload,
}

#[derive(Debug, Serialize, Deserialize)]
struct IPCRequest {
    op: ClientOp,
}

#[derive(Debug, Serialize, Deserialize)]
enum ClientResp<S: AsRef<str>> {
    Ok(Option<S>),
    Error(S),
}

impl<S: AsRef<str> + Serialize> ClientResp<S> {
    async fn write_to<W>(&self, mut writer: W) -> std::io::Result<()>
    where
        W: AsyncWriteExt + std::marker::Unpin,
    {
        writer
            .write_all(serde_json::to_string(self)?.as_bytes())
            .await
    }
}

/*
#[instrument]
async fn delete_mapping(config: &Arc<RwLock<Config<'static>>>) -> Result<()> {
    for mapping in config.read().await.flat_mappings.iter() {
        let (rname, tid, pmap) = mapping;
        let cfg = &config.read().await;
        let mappings = &cfg
            .remotes
            .get(rname)
            .with_context(|| {
                format!("Remote was supposed to exist, but couldn't be found: {rname}")
            })?
            .tunnels
            .iter()
            .nth(*tid)
            .with_context(|| {
                format!("Tunnel was supposed to exist, but couldn't be found: {rname}${tid}")
            })?
            .mappings;
        let (mapping_id, _) = mappings
            .iter()
            .find(|(i, pm)| pm == pmap)
            .context("Port mapping was supposed to exist, but wasn't found")?;
        Ok(mappings.remove(mapping_id))
    }
}
*/

fn delete_mapping_by_flat_id(cfg: Config<'static>, i: usize) -> Result<PortMapping> {
    let (rname, tid, pm) = cfg
        .flat_mappings
        .iter()
        .nth(i)
        .with_context(|| format!("mapping does not exist currently {i}"))?;
    let mappings = &mut cfg
        .remotes
        .get(rname)
        .with_context(|| format!("Remote was expected to exist: {rname}"))?
        .tunnels
        .iter()
        .nth(*tid)
        .with_context(|| format!("Tunnel was expeted to exist {rname}>{tid}"))?
        .mappings;
    if let Some((id, _)) = mappings.iter().find(|(_, m)| pm == m) {
        Ok(mappings.remove(id).unwrap())
    } else {
        bail!(format!(
            "Couldn't find associated port mapping! {rname}>{tid}.{pm}"
        ))
    }
}

#[instrument]
async fn handle_ipc_client(mut conn: UnixStream, app: Arc<AppState>) {
    let config = &app.config;
    let (reader, mut writer) = conn.split();
    let mut reader = BufReader::new(reader);
    let mut buf = String::new();

    if reader.read_line(&mut buf).await.unwrap() == 0 {
        return;
    }

    serde_json::to_string(&*config.read().await).unwrap();

    let resp: ClientResp<_> = match serde_json::from_str::<IPCRequest>(&buf) {
        Ok(req) => match req.op {
            ClientOp::Del(ni) => {
                let cfg = config.write().await;
                for i in ni {}
                todo!()
                //                let cfg = config.write().await;
                //                let km: Vec<String> = ni
                //                    .iter()
                //                    .filter(|i| cfg.mappings.contains_key(i))
                //                    .map(|n| n.to_string())
                //                    .collect();
                //                let kml = km.join(",");
                //                if !km.len() == 0 {
                //                    error!("Bad rule ids when deleting: {}", kml);
                //
                //                    ClientResp::Error(format!("Some route IDs do not exist: {}", kml))
                //                } else {
                //                    for i in ni {
                //                        info!("Removed ruleset {i}!");
                //                    }
                //                    ClientResp::Ok(Some(format!("Removed routes: {kml}")))
                //                }
            }
            ClientOp::Export => {
                let cfg = serde_json::to_string(&*config.read().await).unwrap();
                if let Err(e) = writer.write_all(cfg.as_bytes()).await {
                    error!("Error writing to ipc socket! {e}");
                }
                ClientResp::Ok(None)
            }
            ClientOp::Import => todo!(),
            ClientOp::Add(port_mapping) => {
                let mut cfg = config.write().await;
                let next = cfg.mappings.keys().last().unwrap_or(&0) + 1;
                cfg.mappings.insert(next, port_mapping);
                ClientResp::Ok(None)
            }
            ClientOp::Reload => {
                app.notify.send(()).await.unwrap();
                ClientResp::Ok(None)
            }
        },
        Err(e) => {
            error!("Couldn't parse ipc request: {}", e);
            ClientResp::Error("Bad Request".to_string())
        }
    };

    if let Err(e) = resp.write_to(writer).await {
        error!("Error sending response to client: {e}");
    }
}

#[instrument]
async fn ipc_server(app: Arc<AppState>, sock: UnixListener) {
    loop {
        match sock.accept().await {
            Ok((conn, _)) => {
                tokio::spawn(handle_ipc_client(conn, app.clone()));
            }
            Err(e) => {
                error!("Couldn't accept socket connection: {}", e);
                continue;
            }
        }
    }
}

fn spawn_listeners(cfg: &Config) {
    for (_, _, _mapping) in &cfg.flat_mappings {
        todo!("Spawn threads for every mapping")
    }
}

async fn quic_server(app: Arc<AppState>, mut notify: mpsc::Receiver<()>) {
    let (tx, _rx) = broadcast::channel::<()>(5);

    while notify.recv().await.is_some() {
        let mut active = app.active_config.write().await;
        let cfg = app.config.read().await.clone();
        *active = cfg;
        let a = app.active_config.read().await;
        tx.send(());
        info!("Reloaded config!");
    }
}

#[derive(Debug, clap::Args)]
struct ConnectionAdd {
    #[arg(help = "The name of a registered remote")]
    remote: String,
    #[arg(short, long, help = "Specify a non-default port number")]
    port: u16,
}

#[derive(Subcommand, Debug)]
#[command(
    about = "Operations relating to tunnels",
    long_about = "Multiple port mappings can use the same tunnel. It is recommended to create separate tunnels for high traffic ports"
)]
enum Connection {
    Add(ConnectionAdd),
    List,
    Rm,
    Info,
}

#[derive(Debug, clap::Args)]
struct ServerArgs {
    #[arg(short, long, help = "Causes the server to fork from the shell")]
    daemonize: bool,
    #[arg(short, long, help = "Change the control socket path")]
    socket_path: Option<String>,
}

#[derive(Debug, Clone, PartialEq)]
struct PubKey([u8; 32]);

impl FromStr for PubKey {
    type Err = String;

    fn from_str(s: &str) -> std::result::Result<Self, Self::Err> {
        if s.len() != 43 {
            return Err(format!(
                "Public key must be exactly 43 characters, got {}",
                s.len()
            ));
        }

        let decoded = base64::engine::general_purpose::STANDARD
            .decode(s)
            .map_err(|e| format!("Invalid base64 string: {}", e))?;

        if decoded.len() != 32 {
            return Err(format!(
                "Decoded key must have a length of 32, got {}",
                decoded.len()
            ));
        }

        let mut key = [0u8; 32];
        key.copy_from_slice(&decoded);
        Ok(PubKey(key))
    }
}

impl Serialize for PubKey {
    fn serialize<S>(&self, serializer: S) -> std::result::Result<S::Ok, S::Error>
    where
        S: serde::Serializer,
    {
        serializer.serialize_str(&base64::engine::general_purpose::STANDARD.encode(self.0))
    }
}

impl<'de> Deserialize<'de> for PubKey {
    fn deserialize<D>(deserializer: D) -> std::result::Result<Self, D::Error>
    where
        D: serde::Deserializer<'de>,
    {
        PubKey::from_str(&String::deserialize(deserializer)?).map_err(serde::de::Error::custom)
    }
}

#[derive(Debug, clap::Args)]
struct RemoteAddArgs {
    #[arg(help = "Host ip/domain of a server")]
    host: String,
    #[arg(short, long, help = "Chatter port for the server")]
    port: u8,
    #[arg(short, long, help = "Pubkey for the remote (skips key exchange)")]
    pubkey: PubKey,
}

#[derive(Subcommand, Debug)]
#[command(
    about = "Operations related to remotes (peers)",
    long_about = "Tunnels are created between remotes"
)]
enum RemoteCmd {
    Add(RemoteAddArgs),
    List,
    Rm,
    Info,
}

#[derive(Subcommand, Debug)]
enum Commands {
    #[command(subcommand)]
    Connection(Connection),
    #[command(subcommand)]
    Remote(RemoteCmd),
    License,
    Version,
    Server(ServerArgs),
}

#[derive(Parser, Debug)]
#[command(about = "The tunneling firewall")]
struct Cli {
    #[command(subcommand)]
    command: Commands,
}

static NAME: &str = env!("CARGO_PKG_NAME");
static VERSION: &str = env!("CARGO_PKG_VERSION");
static AUTHOR: &str = env!("CARGO_PKG_AUTHORS");

fn get_socket_path() -> PathBuf {
    if std::env::var("XDG_RUNTIME_DIR").is_ok() {
        std::env::var("XDG_RUNTIME_DIR")
            .map(|dir| PathBuf::from(dir).join("swim.sock"))
            .unwrap()
    } else if std::env::var("SWIM_SYSTEM_DAEMON").is_ok() {
        PathBuf::from("/run/swim.sock")
    } else {
        // Fallback to /tmp
        PathBuf::from("/tmp/swim.sock")
    }
}

#[tokio::main]
async fn main() -> Result<()> {
    let cli = Cli::parse();

    match cli.command {
        Commands::Connection(_connection) => println!("Hello swim!"),
        Commands::License => println!("author(s): {}\n\n{}", AUTHOR, LICENSE),
        Commands::Version => println!("{} {}", NAME, VERSION),
        Commands::Server(cmd) => {
            let sock_path = cmd
                .socket_path
                .map(PathBuf::from)
                .or_else(|| Some(get_socket_path()))
                .unwrap();

            let sock = UnixListener::bind(sock_path).expect("Couldn't bind to unix socket!");
            let (tx, rx) = mpsc::channel(10);

            let config = Config::default();

            let app = Arc::new(AppState {
                config: Arc::new(RwLock::new(config)),
                notify: tx,
                active_config: Arc::new(RwLock::new(Config::default())),
            });

            // TODO: Early config init here

            let ipc_handle = tokio::spawn(ipc_server(app.clone(), sock));
            let quic_handle = tokio::spawn(quic_server(app.clone(), rx));

            ipc_handle.await.unwrap();
            quic_handle.await.unwrap();
        }
        _ => todo!("Not yet implemented!"),
    }
    Ok(())
}
